#----------------------------------------------------------------------------
#
#   MatSheet - Document
#
#----------------------------------------------------------------------------

from numpy import zeros
from GUI import Document, FileType, ApplicationError
import sheetops as ops

mash_type = FileType(name = "MatSheet Document", suffix = "mash", 
    mac_creator = "MaSh", mac_type = "MaSh"
)

#----------------------------------------------------------------------------

block_default_rows = 5
block_default_columns = 3

#----------------------------------------------------------------------------

class BCCoords(object):
    #  Coordinates of a block and cell.
    #
    #  brow   Row number of block
    #  crow   Row number of cell within block
    #  bcol   Column number of block
    #  ccol   Column number of cell within block
    
    def __init__(self, block, cell):
        self.brow, self.bcol = block
        self.crow, self.ccol = cell
    
    def __str__(self):
        return "BCCoords(row=%s,col=%s)" % (
            (self.brow, self.crow), (self.bcol, self.ccol))
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    
    def __ne__(self, other):
        return not self.__eq__(other)


class BCRect(object):
    #  Block and cell coordinates of a rectangle of cells.
    #  Normalized so that left <= right and top <= bottom.
    #
    #  btop      Row number of topmost block
    #  ctop      Row number of topmost cell within block
    #  bleft     Column number of leftmost block
    #  cleft     Column number of leftmost cell within block
    #  bbottom   Row number of bottommost block
    #  cbottom   Row number of bottommost cell within block
    #  bright    Column number of rightmost block
    #  cright    Column number of rightmost cell within block

    def __init__(self, bc1, bc2):
        #  Initialize from block/cell coords of two opposite corners.
        rows = [(bc1.brow, bc1.crow), (bc2.brow, bc2.crow)]
        cols = [(bc1.bcol, bc1.ccol), (bc2.bcol, bc2.ccol)]
        rows.sort()
        cols.sort()
        (self.btop, self.ctop), (self.bbottom, self.cbottom) = rows
        (self.bleft, self.cleft), (self.bright, self.cright) = cols
    
    def __str__(self):
        return "BCRect(top=%s,left=%s,bottom=%s,right=%s)" % (
            (self.btop, self.ctop), (self.bleft, self.cleft),
            (self.bbottom, self.cbottom), (self.bright, self.cright))
    
    def __contains__(self, bc):
        #  True if cell is within rect.
        top = self.btop, self.ctop
        left = self.bleft, self.cleft
        bottom = self.bbottom, self.cbottom
        right = self.bright, self.cright
        row = bc.brow, bc.crow
        col = bc.bcol, bc.ccol
        return top <= row <= bottom and left <= col <= right

#----------------------------------------------------------------------------

class BlockHeader(object):
    #  num_cells   Number of cells wide or high

    def __init__(self, num_cells):
        self.num_cells = num_cells

class RowHeader(BlockHeader):
    pass

class ColumnHeader(BlockHeader):
    #  width   Display width of cell in pixels
    
    cell_width = 50

#----------------------------------------------------------------------------

class Block(object):
    #  data   array[row, col]
    
    left_margin = 1
    right_margin = 1
    align = 'r'

    def __init__(self, num_rows = 5, num_columns = 3):
        self.data = zeros((num_rows, num_columns))
    
    def get_num_rows(self):
        return self.data.shape[0]
    
    def get_num_columns(self):
        return self.data.shape[1]
    
    num_rows = property(get_num_rows)
    num_columns = property(get_num_columns)

    def text_to_value(self, text):
        try:
            return float(text)
        except ValueError:
            raise ApplicationError("Invalid numeric value.")

#----------------------------------------------------------------------------

class SheetDoc(Document):
    #  row_headers      RowHeader[num_rows]
    #  column_headers   ColumnHeader[num_columns]
    #  blocks           Block[num_rows][num_columns]
    #  selection        BCRect or None
    #  editing_layout   boolean
    #  undo_stack       [SheetOp]
    #  redo_stack       [SheetOp]

    def __init__(self):
        Document.__init__(self, file_type = mash_type)
        b = Block()
        self.row_headers = [RowHeader(b.num_rows)]
        self.column_headers = [ColumnHeader(b.num_columns)]
        self.blocks = [[b]]
        self.selection = None
        self.editing_layout = False
        self.undo_stack = []
        self.redo_stack = []

    def setup_menus(self, m):
        m.undo_cmd.enabled = bool(self.undo_stack)
        m.redo_cmd.enabled = bool(self.redo_stack)

    def undo_cmd(self):
        op = self.undo_stack.pop()
        self.redo_stack.append(op.do(self))
    
    def redo_cmd(self):
        self.do(self.redo_stack.pop())

    def get_num_rows(self):
        return len(self.row_headers)
    
    def get_num_columns(self):
        return len(self.column_headers)
    
    num_rows = property(get_num_rows)
    num_columns = property(get_num_columns)

    def sync(self):
        self.notify_views('model_sync')

    def toggle_editing_layout(self):
        self.editing_layout = not self.editing_layout
        self.notify_views()

    def set_selection(self, bcrect):
        self.selection = bcrect
        self.notify_views()

    def get_cell_value(self, bc):
        return self.blocks[bc.brow][bc.bcol].data[bc.crow, bc.ccol]
    
    def set_cell_value(self, bc, value):
        self.blocks[bc.brow][bc.bcol].data[bc.crow, bc.ccol] = value
        self.notify_views()
    
    def do(self, op):
        self.undo_stack.append(op.do(self))
        self.changed()
    
    def do_change_cell(self, bc, value):
        if self.get_cell_value(bc) != value:
            self.do(ops.ChangeCell(bc, value))
